'''
Created: Aug 4, 2013
Last modified: Aug 13, 2013
@author Flaise Saffron
'''

import pygame
import math
import sys
import random
import os


######################################################################## engine

### Not used in this app anymore but still a useful library function
def get_stick(joy, axisX, axisY):
	state = ( joy.get_axis(axisX), joy.get_axis(axisY) )
	
	magSq = magnitudeSq2d(state)
	
	### The MadCatz controller reports a vector magnitude of >1 when the sticks
	### are tilted towards the corners, which seems to be because of a dead
	### zone around the outside. Normalizing the vector whenever its magnitude
	### is >1 seems to fix this.
	if magSq > 1:
		return normalize2d(state)
	
	dead_zone_radius = .2
	
	### Circular dead zone
	if magSq < dead_zone_radius ** 2:
		return (0, 0)
	
	### Scale usable input space so it goes from 0->1 instead of abruptly
	### jumping to the hardware's representation of the edge of the dead zone.
	mag = math.sqrt(magSq)
	max_mag = 1 - dead_zone_radius
	new_mag = (mag - dead_zone_radius) / max_mag
	
	return mag_scale_to_2d(state, new_mag)

### Tampering with the pygame Vector classes to add the necessary functionality
### didn't seem worth it; these few functions are all that are needed anyway
def sum2d(a, b):
	return (a[0] + b[0], a[1] + b[1])

def dif2d(a, b):
	return (a[0] - b[0], a[1] - b[1])

def sum3d(a, b):
	return (a[0] + b[0], a[1] + b[1], a[2] + b[2])

def magnitudeSq2d(vec):
	return vec[0] ** 2 + vec[1] ** 2

def magnitude2d(vec):
	return math.sqrt(magnitudeSq2d(vec))

def normalize2d(vec):
	return mag_scale_2d(vec, 1 / magnitude2d(vec))

def int2d(vec):
	return (int(vec[0]), int(vec[1]))

def mag_scale_2d(vec, scalar):
	return (vec[0] * scalar, vec[1] * scalar)

def mag_scale_to_2d(vec, scalar):
	if not math.isfinite(scalar):
		raise Exception() ### because this was probably not intended
	return mag_scale_2d(vec, scalar / magnitude2d(vec))
	### Just let it throw when the vector has a magnitude of zero; there's no
	### real way to klodge that into a usable number because it would require
	### a direction.
	### A real programming language would allow this function to have
	### preconditions.

def flipY2d(vec):
	return (vec[0], -vec[1])

def flip2d(vec):
	return (-vec[0], -vec[1])

def scale2d(vec, scalar):
	return (vec[0] * scalar, vec[1] * scalar)

def basic_product_2d(a, b):
	return (a[0] * b[0], a[1] * b[1])

def negate2d(vec):
	return (-vec[0], -vec[1])

def negate3d(vec):
	return (-vec[0], -vec[1], -vec[2])


EVENT_DRAW = 100

event_handlers = [] ### contains: (predicate:function, callback:function)
next_event_handlers = []
def proc_event(event):
	global event_handlers, next_event_handlers
	
	event_handlers = next_event_handlers
	next_event_handlers = list(event_handlers)
	for handler in event_handlers:
		if handler[0](event):
			handler[1](event)
			
def proc_events():
	for event in pygame.event.get():
		proc_event(event)
		
def event_loop():
	while True:
		proc_events()
		proc_event(pygame.event.Event(EVENT_DRAW))
		pygame.display.flip()

def on_event(predicate, callback):
	### TODO: take priorty as parameter so screen flip and hud drawing can be
	### registered in any order
	
	### TODO: what about attaching events to other events?
	
	removed = False
	def invoke(event):
		if removed: return
		
		callback(event)
	
	def remove(): ### TODO: Use linked list
		nonlocal removed
		if removed: return
		removed = True
		
		next_event_handlers.remove(handler)
	
	handler = (predicate, invoke)
	next_event_handlers.append(handler)
	return remove

def on_event_once(predicate, callback):
	stop = on_event(predicate, lambda a: (stop(), callback(a)))


def render_text(text, position, anchor, dest_surface, font, color_fill, color_outline):
	surface = font.render(text, True, color_outline)
	position = sum2d(position, negate2d(basic_product_2d(surface.get_size(), anchor)))
	
	dest_surface.blit(surface, sum2d(position, (2, 0)))
	dest_surface.blit(surface, sum2d(position, (-2, 0)))
	dest_surface.blit(surface, sum2d(position, (0, 2)))
	dest_surface.blit(surface, sum2d(position, (0, -2)))
	
	surface = font.render(text, True, color_fill)
	dest_surface.blit(surface, position)
	
	return pygame.Rect(position, surface.get_size())



########################################################################### app

pygame.init()
surface_screen = pygame.display.set_mode((1024, 768))
pygame.display.set_caption('Squish the Bugs v0.1')

joysticks = [pygame.joystick.Joystick(x) for x in range(pygame.joystick.get_count())]
for joy in joysticks:
	joy.init()

color_BG = pygame.Color(220, 220, 255)
color_HUD = pygame.Color(255, 240, 240)
color_HUD_outline = pygame.Color(60, 30, 30)

class AppEvent:
	(
		INPUT_DIRECTION, ### kwargs: direction
		BLOCK_MOVED, ### kwargs: block
		BUG_SQUISHED, ### kwargs: block
		RESTART ### void
	) = range(1000, 1004)
	
class Direction:
	NORTH, EAST, SOUTH, WEST = range(4)

on_event(
	lambda a: a.type == pygame.KEYDOWN and a.key == pygame.K_UP,
	lambda a: proc_event(pygame.event.Event(AppEvent.INPUT_DIRECTION, direction=Direction.NORTH))
)
on_event(
	lambda a: a.type == pygame.KEYDOWN and a.key == pygame.K_DOWN,
	lambda a: proc_event(pygame.event.Event(AppEvent.INPUT_DIRECTION, direction=Direction.SOUTH))
)
on_event(
	lambda a: a.type == pygame.KEYDOWN and a.key == pygame.K_RIGHT,
	lambda a: proc_event(pygame.event.Event(AppEvent.INPUT_DIRECTION, direction=Direction.EAST))
)
on_event(
	lambda a: a.type == pygame.KEYDOWN and a.key == pygame.K_LEFT,
	lambda a: proc_event(pygame.event.Event(AppEvent.INPUT_DIRECTION, direction=Direction.WEST))
)

previous_hat = (0, 0)
def hat_to_direction4(event):
	global camera, previous_hat
	if event.value[0] != previous_hat[0]:
		if event.value[0] == 1:
			proc_event(pygame.event.Event(AppEvent.INPUT_DIRECTION, direction=Direction.EAST))
		elif event.value[0] == -1:
			proc_event(pygame.event.Event(AppEvent.INPUT_DIRECTION, direction=Direction.WEST))
	if event.value[1] != previous_hat[1]:
		if event.value[1] == 1:
			proc_event(pygame.event.Event(AppEvent.INPUT_DIRECTION, direction=Direction.NORTH))
		elif event.value[1] == -1:
			proc_event(pygame.event.Event(AppEvent.INPUT_DIRECTION, direction=Direction.SOUTH))
	previous_hat = event.value
on_event(lambda a: a.type == pygame.JOYHATMOTION, lambda a: hat_to_direction4(a))


on_event(lambda a: a.type == EVENT_DRAW, lambda a: surface_screen.fill(color_BG))
on_event(lambda a: a.type == pygame.QUIT, lambda a: sys.exit())
on_event(lambda a: a.type == pygame.KEYDOWN and a.key == pygame.K_ESCAPE, lambda a: sys.exit())


camera = ((0, 0, 0), 7) # ((x, y, z), view_height) ~~~ width depends on screen

def move_camera_to(dest):
	if len(dest) != 3: raise Exception() 
	global camera
	camera = (dest, camera[1])


TILE_W = 1
TILE_L = .8
TILE_H = .4

### tile: (surface, (reg_x, reg_y), reg_width, is_filler, relative_layer, hud)
tile_gray = (pygame.image.load(os.path.join('images', 'Plain Block.png')).convert_alpha(), (50, 90), 100, True, 0, False)
tile_dirt = (pygame.image.load(os.path.join('images', 'Dirt Block.png')).convert_alpha(), (50, 90), 100, True, 0, False)
tile_grass = (pygame.image.load(os.path.join('images', 'Grass Block.png')).convert_alpha(), (50, 90), 100, True, 0, False)
tile_water = (pygame.image.load(os.path.join('images', 'Water Block.png')).convert_alpha(), (50, 90), 100, True, -1, False)
tile_rock = (pygame.image.load(os.path.join('images', 'Rock.png')).convert_alpha(), (50, 90), 100, False, 0, False)
tile_tree = (pygame.image.load(os.path.join('images', 'Tree Tall.png')).convert_alpha(), (50, 90), 100, False, 0, False)
tile_player = (pygame.image.load(os.path.join('images', 'Character Cat Girl.png')).convert_alpha(), (50, 90), 100, False, 0, False)
tile_player_in_water = (pygame.image.load(os.path.join('images', 'Character Cat Girl In Water.png')).convert_alpha(), (50, 90), 100, False, 0, False)
tile_bug = (pygame.image.load(os.path.join('images', 'Enemy Bug.png')).convert_alpha(), (50, 90), 100, False, 0, False)
tile_bug_dead = (pygame.image.load(os.path.join('images', 'Enemy Bug Squished.png')).convert_alpha(), (50, 90), 100, False, -1, False)
tile_speech = (pygame.image.load(os.path.join('images', 'SpeechBubble.png')).convert_alpha(), (-35, 190), 100, False, 0, True)
tile_wood = (pygame.image.load(os.path.join('images', 'Wood Block.png')).convert_alpha(), (50, 90), 100, True, 0, False)
tile_question = (pygame.image.load(os.path.join('images', 'Question.png')).convert_alpha(), (-35, 190), 100, False, 1, True) ### Uses font Palatino Linotype
tile_bush_0 = (pygame.image.load(os.path.join('images', 'Tree Short.png')).convert_alpha(), (50, 90), 100, False, 0, False)
tile_bush_1 = (pygame.image.load(os.path.join('images', 'Tree Ugly.png')).convert_alpha(), (50, 90), 100, False, 0, False)

shadow_e = (pygame.image.load(os.path.join('images', 'Shadow East.png')), (50, 90), 100)
shadow_w = (pygame.image.load(os.path.join('images', 'Shadow West.png')), (50, 90), 100)
shadow_s = (pygame.image.load(os.path.join('images', 'Shadow South.png')), (50, 90), 100)
shadow_n = (pygame.image.load(os.path.join('images', 'Shadow North.png')), (50, 90), 100)
shadow_side_w = (pygame.image.load(os.path.join('images', 'Shadow Side West.png')), (50, 90), 100)
shadow_se = (pygame.image.load(os.path.join('images', 'Shadow South East.png')), (50, 90), 100)
shadow_nw = (pygame.image.load(os.path.join('images', 'Shadow North West.png')), (50, 90), 100)
shadow_sw = (pygame.image.load(os.path.join('images', 'Shadow South West.png')), (50, 90), 100)
shadow_ne = (pygame.image.load(os.path.join('images', 'Shadow North East.png')), (50, 90), 100)


class BlockType:
	PLAYER, BUG, ROCK, TREE, FILLER, BUG_DEAD, SPEECH, WATER = range(8)

blocks = [] ### contains: ((x, y, z), tile, BlockType)
position_to_blocklist = {} ### contains: {(x, y, z): [blocks]}


def clear_blocks():
	global blocks
	blocks = [] ### Not sure why List doesn't have a clear method...
	
	position_to_blocklist.clear()
on_event(lambda a: a.type == AppEvent.RESTART, lambda a: clear_blocks())

def update_blocks():
	blocks.sort(key=lambda a: a[0][2] * 10000 + a[0][1] + a[1][4] + (100000000 if a[1][5] else 0))
	position_to_blocklist.clear()
	for block in blocks:
		position = block[0]
		if position not in position_to_blocklist:
			position_to_blocklist[position] = []
		position_to_blocklist[position].append(block)

def remove_block(block):
	position_to_blocklist[block[0]].remove(block)
	if len(position_to_blocklist[block[0]]) == 0:
		del position_to_blocklist[block[0]]
	blocks.remove(block)
	proc_event(pygame.event.Event(AppEvent.BLOCK_MOVED, block=block, dest=None))
	
def is_block_at(position):
	return position in position_to_blocklist and len(position_to_blocklist[position]) > 0

def move_block_to(block, dest):
	block[0] = dest
	update_blocks()
	proc_event(pygame.event.Event(AppEvent.BLOCK_MOVED, block=block, dest=dest))
	
def move_block(block, delta):
	move_block_to(block, sum3d(block[0], delta))

def get_block_types_at(position):
	for block in get_blocks_at(position):
		yield block[2]
		
def get_blocks_at(position):
	if position in position_to_blocklist:
		for block in position_to_blocklist[position]:
			yield block
			
def is_block_type_at(position, block_type):
	return block_type in get_block_types_at(position)

def make_block(position, tile, block_type):
	if len(position) != 3: raise Exception()
	block = [position, tile, block_type]
	blocks.append(block)
	update_blocks()
	return block



########## Player

block_player = None

def move_player(block, delta):
	def is_obstacle_type(block_type):
		return block_type in (BlockType.BUG, BlockType.ROCK, BlockType.TREE, BlockType.FILLER)
	
	def is_obstacle_at(position):
		return len([x for x in get_block_types_at(position) if is_obstacle_type(x)]) > 0
	
	def on_fail():
		bubble = make_block(block[0], tile_speech, BlockType.SPEECH)
		question = make_block(block[0], tile_question, BlockType.SPEECH)
		
		def remove(event):
			if event.type == AppEvent.RESTART:
				return
			remove_block(bubble)
			remove_block(question)
		on_event_once(lambda a: a.type == AppEvent.INPUT_DIRECTION or a.type == AppEvent.RESTART, remove)
	
	dest = sum3d(block[0], delta)
	above = sum3d(dest, (0, 1, 0))
	
	if is_obstacle_at(dest):
		if is_obstacle_at(above):
			on_fail()
			return
		for block_type in get_block_types_at(dest):
			if block_type == BlockType.FILLER:
				move_block(block, sum3d(delta, (0, 1, 0)))
				return
		on_fail()
		return
	else:
		for i in range(1, 5):
			below = sum3d(dest, (0, -i, 0))
			
			for other_block in get_blocks_at(below):
				block_type = other_block[2]
				
				if block_type in (BlockType.FILLER, BlockType.ROCK):
					move_block(block, sum3d(delta, (0, 1 - i, 0)))
					return
				elif block_type == BlockType.BUG:
					squish_bug(other_block)
					move_block(block, sum3d(delta, (0, -i, 0)))
					return
				elif is_obstacle_type(block_type):
					on_fail()
					return
				else:
					continue
		on_fail()
	
def make_player(position):
	global block_player
	block_player = make_block(position, tile_player, BlockType.PLAYER)
	
	def center_camera_on_player():
		move_camera_to(sum3d(block_player[0], (0, -1, 0)))
	center_camera_on_player()
	
	def update_tile():
		if is_block_type_at(block_player[0], BlockType.WATER):
			block_player[1] = tile_player_in_water
		else:
			block_player[1] = tile_player
	update_tile()
	
	
	on_event(lambda a: a.type == AppEvent.BLOCK_MOVED and a.block == block_player, lambda a: (
		center_camera_on_player(),
		update_tile()
	))


########## Bug

num_bugs = 0
def restart_bugs():
	global num_bugs
	num_bugs = 0
on_event(lambda a: a.type == AppEvent.RESTART, lambda a: restart_bugs())

def squish_bug(block):
	global num_bugs
	num_bugs -= 1
	
	block[1] = tile_bug_dead
	block[2] = BlockType.BUG_DEAD
	
	proc_event(pygame.event.Event(AppEvent.BUG_SQUISHED, block=block))

def can_bug_move(block, delta):
	if len(list(get_blocks_at(sum3d(block[0], (0, 1, 0))))) > 0: # something on top, usually another bug
		return False # let other bug move off first so odd behavior doesn't commence
	
	def is_obstacle(block_type):
		return block_type in (BlockType.PLAYER, BlockType.BUG, BlockType.ROCK, BlockType.TREE, BlockType.FILLER)
	def is_obstacle_at(position):
		return len([x for x in get_block_types_at(position) if is_obstacle(x)]) > 0
	
	dest = sum3d(block[0], delta)
	
	if is_obstacle_at(dest):
		return False
	else:
		below = sum3d(dest, (0, -1, 0))
		
		for other_block in get_blocks_at(below):
			block_type = other_block[2]
			
			if block_type in (BlockType.FILLER, BlockType.ROCK, BlockType.BUG):
				return True
			elif not is_obstacle(block_type):
				continue
			return False
		
def move_bug(block, delta):
	if can_bug_move(block, delta):
		move_block(block, delta)
		
def make_bug(position):
	global num_bugs
	num_bugs += 1
	
	last_move = None
	def do_ai(event):
		nonlocal last_move
		
		if block[2] == BlockType.BUG_DEAD: return
		
		deltas = [(1, 0, 0), (-1, 0, 0), (0, 0, 1), (0, 0, -1)]
		
		reverse_move = (None if last_move == None else negate3d(last_move))
		if last_move != None:
			deltas.remove(reverse_move) ### avoid about-face when possible 
		
		options = [x for x in deltas if can_bug_move(block, x)]
		if len(options) > 0:
			option = random.choice(options)
			move_bug(block, option)
			last_move = option
		else:
			### attempt about-face when all else fails
			if reverse_move != None and can_bug_move(block, reverse_move):
				move_bug(block, reverse_move)
				last_move = reverse_move
	
	block = make_block(position, tile_bug, BlockType.BUG)
	on_event(lambda a: a.type == AppEvent.BLOCK_MOVED and a.block == block_player, do_ai)
	### TODO: Remove event binding when bug-death event fires
	



def make_grass(position):
	make_block(position, tile_grass, BlockType.FILLER)
def make_stone(position):
	make_block(position, tile_gray, BlockType.FILLER)
def make_boulder(position):
	make_block(position, tile_rock, BlockType.ROCK)
def make_tree(position):
	make_block(position, tile_tree, BlockType.TREE)
def make_bush_0(position):
	make_block(position, tile_bush_0, BlockType.TREE)
def make_bush_1(position):
	make_block(position, tile_bush_1, BlockType.TREE)
def make_wood_floor(position):
	make_block(position, tile_wood, BlockType.FILLER)
def make_water(position):
	make_block(position, tile_water, BlockType.WATER)
def make_dirt(position):
	make_block(position, tile_dirt, BlockType.FILLER)


scale_cache = {} ### { surface: ((x, y), result_surface) }
def scale_surface_to(surface, dest_size):
	if not (surface in scale_cache and scale_cache[surface][0] == dest_size):
		scale_cache[surface] = (dest_size, pygame.transform.smoothscale(surface, dest_size))
	return scale_cache[surface][1]

def draw_tile(tile, position):
	tile_surface = tile[0]
	tile_reg_width = tile[2]
	
	scale_world2screen = surface_screen.get_height() / (camera[1] * TILE_L)
	scale_image2tile = tile_surface.get_width() / tile_reg_width
	
	dest_w = scale_world2screen * scale_image2tile
	dest_h = dest_w * tile_surface.get_height() / tile_surface.get_width()
	surface_screen.blit(
		scale_surface_to(tile_surface, int2d((dest_w, dest_h))),
		sum2d(
			### positioning before camera motion
			dif2d(
				scale2d(
					(
						position[0] * TILE_W,
						position[2] * TILE_L - position[1] * TILE_H
					),
					scale_world2screen
				),
				scale2d(tile[1], scale_world2screen / tile_reg_width)
			),
			
			sum2d(
				### camera position
				scale2d(
					(TILE_W * -camera[0][0], TILE_H * camera[0][1] - TILE_L * camera[0][2]), 
					scale_world2screen
				),
				(surface_screen.get_width() / 2, surface_screen.get_height() / 2)
			)
		)
	)

def draw_environment(event):
	def is_filler_at(position):
		for block in get_blocks_at(position):
			if block[1][3]:
				return True
		return False
	
	for block in blocks:
		tile = block[1]
		position = block[0]
		draw_tile(tile, position)
		if tile[3]:
			### surface shadows
			if not is_filler_at(sum3d(position, (0, 1, 0))):
				if is_filler_at(sum3d(position, (1, 1, 0))):
					draw_tile(shadow_e, position)
				if is_filler_at(sum3d(position, (-1, 1, 0))):
					draw_tile(shadow_w, position)
				if is_filler_at(sum3d(position, (0, 1, 1))):
					draw_tile(shadow_s, position)
				if is_filler_at(sum3d(position, (0, 1, -1))):
					draw_tile(shadow_n, position)
				
				### corners
				if not is_filler_at(sum3d(position, (0, 1, 1))):
					if is_filler_at(sum3d(position, (1, 1, 1))) and not is_filler_at(sum3d(position, (1, 1, 0))):
						draw_tile(shadow_se, position)
					if is_filler_at(sum3d(position, (-1, 1, 1))) and not is_filler_at(sum3d(position, (-1, 1, 0))):
						draw_tile(shadow_sw, position)
				if is_filler_at(sum3d(position, (-1, 1, -1))) and not is_filler_at(sum3d(position, (0, 1, -1))) and not is_filler_at(sum3d(position, (-1, 1, 0))):
					draw_tile(shadow_nw, position)
				if is_filler_at(sum3d(position, (1, 1, -1))) and not is_filler_at(sum3d(position, (0, 1, -1))) and not is_filler_at(sum3d(position, (1, 1, 0))):
					draw_tile(shadow_ne, position)
			
			### front shadows
			if not is_filler_at(sum3d(position, (0, 0, 1))):
				if is_filler_at(sum3d(position, (-1, 0, 1))):
					draw_tile(shadow_side_w, position)
					
on_event(lambda a: a.type == EVENT_DRAW, draw_environment)

on_event(lambda a: a.type == AppEvent.INPUT_DIRECTION and a.direction == Direction.NORTH, lambda a: move_player(block_player, (0, 0, -1)))
on_event(lambda a: a.type == AppEvent.INPUT_DIRECTION and a.direction == Direction.SOUTH, lambda a: move_player(block_player, (0, 0, 1)))
on_event(lambda a: a.type == AppEvent.INPUT_DIRECTION and a.direction == Direction.EAST, lambda a: move_player(block_player, (1, 0, 0)))
on_event(lambda a: a.type == AppEvent.INPUT_DIRECTION and a.direction == Direction.WEST, lambda a: move_player(block_player, (-1, 0, 0)))

def ms_to_hms_string(ms):
	seconds = int(ms / 1000)
	
	minutes = int(seconds / 60)
	seconds %= 60
	
	hours = int(minutes / 60)
	minutes %= 60
	
	return ('{0:02d}'.format(hours) + ':' if hours > 0 else '') + '{0:02d}'.format(minutes) + ':' + '{0:02d}'.format(seconds)

fps_clock = pygame.time.Clock()
gui_font_large = pygame.font.SysFont('Verdana', 32)
gui_font_medium = pygame.font.SysFont('Verdana', 20)
gui_font_small = pygame.font.SysFont('Verdana', 14)

puzzle_start = pygame.time.get_ticks()
puzzle_end = None

def restart_puzzle_time():
	global puzzle_start
	global puzzle_end
	puzzle_start = pygame.time.get_ticks()
	puzzle_end = None
on_event(lambda a: a.type == AppEvent.RESTART, lambda a: restart_puzzle_time())

def get_puzzle_duration():
	if puzzle_end != None:
		return puzzle_end - puzzle_start
	return pygame.time.get_ticks() - puzzle_start

def end_puzzle():
	global puzzle_end
	puzzle_end = pygame.time.get_ticks()
on_event(lambda a: a.type == AppEvent.BUG_SQUISHED and num_bugs == 0, lambda a: end_puzzle())

def draw_hud(event):
	fps_clock.tick()
	
	### Keep text away from edges so it's inside the title safe zone
	render_text('FPS: ' + str(int(fps_clock.get_fps())), sum2d((45, -45), surface_screen.get_rect().bottomleft), (0, 1), surface_screen, gui_font_small, color_HUD, color_HUD_outline)
	
	instruction_rect_0 = render_text('Squish all the bugs.', sum2d((-45, 45), surface_screen.get_rect().topright), (1, 0), surface_screen, gui_font_medium, color_HUD, color_HUD_outline)
	instruction_rect_1 = render_text('Use the arrow keys or d-pad to move.', sum2d((0, 10), instruction_rect_0.bottomright), (1, 0), surface_screen, gui_font_medium, color_HUD, color_HUD_outline)
	instruction_rect_2 = render_text('If the puzzle becomes unsolvable, press', instruction_rect_1.bottomright, (1, 0), surface_screen, gui_font_medium, color_HUD, color_HUD_outline)
	render_text('Space or any gamepad button to start over.', instruction_rect_2.bottomright, (1, 0), surface_screen, gui_font_medium, color_HUD, color_HUD_outline)
	
	if num_bugs > 0:
		render_text('Elapsed: ' + ms_to_hms_string(get_puzzle_duration()), (45, 45), (0, 0), surface_screen, gui_font_large, color_HUD, color_HUD_outline)
	else:
		render_text('Solved in: ' + ms_to_hms_string(get_puzzle_duration()), (45, 45), (0, 0), surface_screen, gui_font_large, color_HUD, color_HUD_outline)
on_event(lambda a: a.type == EVENT_DRAW, draw_hud)



#################################################################### unit tests

assert mag_scale_to_2d((1, 0), 2) == (2, 0)
assert mag_scale_to_2d((-2, 0), .5) == (-.5, 0)
assert mag_scale_to_2d((1, 1), 1) == (1 / math.sqrt(2), 1 / math.sqrt(2))



######################################################################### begin

### obviously not the best map loading solution in a large app, but whatever
import map_0

def load_map():
	clear_blocks()
	map_0.make(make_player, make_bug, make_grass, make_stone, make_boulder, make_tree, make_bush_0, make_bush_1, make_wood_floor, make_water, make_dirt)
load_map()

on_event(lambda a: a.type == AppEvent.RESTART, lambda a: load_map())

on_event(
	lambda a: (a.type == pygame.KEYDOWN and a.key == pygame.K_SPACE) or a.type == pygame.JOYBUTTONDOWN,
	lambda a: proc_event(pygame.event.Event(AppEvent.RESTART))
)


event_loop()
